import json
from typing import Any


def dump(file: str, obj: Any) -> None:
    """Serialize obj as a JSON formatted stream to file.

    Args:
        file (str): File path.
        obj (Any): Object to be serialized.
    """
    with open(file, 'w', encoding='utf-8') as file:
        json.dump(obj, file, ensure_ascii=False,
                  indent=2, sort_keys=False)


def load(file: str) -> Any:
    """Deserialize file to a Python object.

    Args:
        file (str): File path.

    Returns:
        Any: Python object.
    """
    with open(file, 'r', encoding='utf-8') as file:
        return json.load(file)



import requests

from .context import Context
from .env import CLUEDIN_REQUEST_TIMEOUT_IN_SECONDS


def get_entity_blob(context: Context, entity_id: str) -> str:
    """Get entity blob.

    Args:
        context (Context): Context object.
        entity_id (str): Entity ID.

    Returns:
        dict: XML response - entity blob.
    """

    headers = {
        'Authorization': f'Bearer {context.access_token}'
    }

    params = {
        'id': entity_id
    }

    response = requests.get(
        url=f'{context.api_url}/admin/entity/blob',
        headers=headers,
        params=params,
        timeout=CLUEDIN_REQUEST_TIMEOUT_IN_SECONDS,
        verify=context.verify_tls)

    if not response.ok:
        response.raise_for_status()

    return response.text


def get_entity_as_clue(context: Context, entity_id: str) -> str:
    """Get entity as clue.

    Args:
        context (Context): Context object.
        entity_id (str): Entity ID.

    Returns:
        dict: XML response - entity clue.
    """

    headers = {
        'Authorization': f'Bearer {context.access_token}'
    }

    params = {
        'id': entity_id
    }

    response = requests.get(
        url=f'{context.api_url}/admin/entity/clue',
        headers=headers,
        params=params,
        timeout=CLUEDIN_REQUEST_TIMEOUT_IN_SECONDS,
        verify=context.verify_tls)

    if not response.ok:
        response.raise_for_status()

    return response.text



from datetime import datetime


class Rule:
    """
    Represents a rule object.

    Args:
        rule (dict): The rule dictionary containing the rule information.

    Attributes:
        id (str): The ID of the rule.
        field (str): The normalized field of the rule.
        operator_id (str): The ID of the operator.
        type (str): The type of the rule.
        value (str or int or float or list): The value of the rule.

    Methods:
        normalize_field: Normalizes the field value.
        get_value: Returns the typecasted value of the rule.
        typecast_value: Typecasts the given value based on the rule type.
    """

    def __init__(self, rule) -> None:
        self.id = rule['id']
        self.field = rule['field']
        self.operator_id = rule['operator']
        self.type = rule['type']
        self.value = rule['value']

    def get_value(self):
        """
        Returns the value of the rule.

        If the value is a list, it will be typecasted and returned as a list.
        If the value is a single element, it will be typecasted and returned as is.
        If the value is neither a list nor a single element, it will be typecasted and returned.

        Returns:
            The typecasted value of the rule.
        """
        if isinstance(self.value, list):
            if len(self.value) == 1:
                return self.typecast_value(self.value[0])
            return [self.typecast_value(x) for x in self.value]
        return self.typecast_value(self.value)

    def typecast_value(self, value_to_cast):
        """
        Typecasts the given value based on the rule's type.

        Args:
            value_to_cast: The value to be typecasted.

        Returns:
            The typecasted value, or None if the value cannot be typecasted.

        """
        if value_to_cast is None:
            return None

        try:
            if self.type == 'string':
                return str(value_to_cast)
            if self.type == 'integer':
                return int(value_to_cast)
            if self.type == 'double':
                return float(value_to_cast)
            if self.type == 'date' or self.type == 'datetime':
                if isinstance(value_to_cast, str):
                    return datetime.strptime(value_to_cast, "%Y-%m-%dT%H:%M:%S.%fZ")
                return value_to_cast
            if self.type == 'boolean' and isinstance(value_to_cast, str):
                return value_to_cast.lower() in ['true', '1']
            return value_to_cast
        except ValueError:
            return None



import json
from typing import Any

from .operators import default_get_operator, pandas_get_operator
from .rule import Rule
from .rule_group import RuleGroup


def default_get_property_name(field: str) -> str:
    """
    Returns the property name for a given field.

    Args:
        field (str): The field name.

    Returns:
        str: The property name.

    """
    if field is None:
        return None
    if field == 'EntityType':
        field = 'entityType'
    return field.replace('Properties[', '').replace(']', '')


def default_get_value(field: str, obj: dict) -> Any:
    """
    Get the value of a field from a dictionary object.

    Args:
        field (str): The field name.
        obj (dict): The dictionary object.

    Returns:
        Any: The value of the field in the dictionary object.
    """
    return obj.get(field)


class Evaluator:
    """
    The Evaluator class is responsible for evaluating rules against objects and returning
    a list of objects that match the rules.

    Args:
        rule_set (str or list): The rule set to be evaluated. It can be either a JSON string
            representing the rule set or a list of rule objects.
        get_operator (function, optional): A function that returns the operator based on the
            operator ID. Defaults to default_get_operator.
        get_property_name (function, optional): A function that returns the property name based
            on the field ID. Defaults to default_get_property_name.
        get_value (function, optional): A function that returns the value based on the property
            name and the object. Defaults to default_get_value.

    Attributes:
        rule_group (RuleGroup): The rule group representing the rule set.

    Methods:
        get_matching_objects(objects): Returns a list of objects that match the rules.
        object_matches_rules(obj): Checks if the given object matches the rules.
        explain(): Generates a pandas query string based on the rules.

    Raises:
        ValueError: If the condition of the rule group is invalid.
    """
    def __init__(self,
                 rule_set,
                 get_operator=default_get_operator,
                 get_property_name=default_get_property_name,
                 get_value=default_get_value):
        self.get_operator = get_operator
        self.get_property_name = get_property_name
        self.get_value = get_value
        if isinstance(rule_set, str):
            self.rule_group = RuleGroup(json.loads(rule_set))
        else:
            self.rule_group = RuleGroup(rule_set)

    def get_matching_objects(self, objects):
        """
        Returns a list of objects that match the rules.

        Args:
            objects (list): The list of objects to be evaluated.

        Returns:
            list: A list of objects that match the rules.
        """
        return list(filter(self.object_matches_rules, objects))

    def object_matches_rules(self, obj):
        """
        Checks if the given object matches the rules defined in the rule group.

        Args:
            obj: The object to be evaluated against the rules.

        Returns:
            bool: True if the object matches the rules, False otherwise.
        """
        return self.__evaluate_rule_group(self.rule_group, obj)

    def __evaluate_rule_group(self, rule_group, obj):
        """
        Evaluates a rule group based on the given condition and rules.

        Args:
            rule_group (RuleGroup): The rule group to evaluate.
            obj (object): The object to evaluate against the rules.

        Returns:
            bool: True if the rule group evaluates to True, False otherwise.

        Raises:
            ValueError: If the condition of the rule group is invalid.
        """
        if rule_group.condition == 'AND':
            return all(map(lambda x: self.__evaluate_rule_object(x, obj), rule_group.rules))
        if rule_group.condition == 'OR':
            return any(map(lambda x: self.__evaluate_rule_object(x, obj), rule_group.rules))
        raise ValueError(f"Invalid condition: {rule_group.condition}")

    def __evaluate_rule(self, rule: dict, obj: dict) -> bool:
        """
        Evaluates a rule against an object.

        Args:
            rule (Rule): The rule to evaluate.
            obj (Object): The object to evaluate the rule against.

        Returns:
            bool: True if the rule is satisfied, False otherwise.
        """
        return self.get_operator(rule.operator_id)(
            # left - value from the evaluated object
            rule.typecast_value(
                self.get_value(
                    self.get_property_name(rule.field),
                    obj)),
            # right - value from the rule
            rule.get_value(),
            obj)

    def __evaluate_rule_object(self, rule_object: dict, obj) -> bool:
        """
        Evaluates a rule object against the given object.

        Args:
            rule_object (dict): The rule object to evaluate.
            obj: The object to evaluate the rule against.

        Returns:
            The result of the evaluation.
        """
        if 'rules' in rule_object and len(rule_object['rules']) > 0:
            return self.__evaluate_rule_group(RuleGroup(rule_object), obj)
        return self.__evaluate_rule(Rule(rule_object), obj)

    def __explain_rule_object(self, rule_object):
        """
        Generates a pandas query string based on the given rule.

        Args:
            rule (Rule): The rule to generate the query for.

        Returns:
            str: The pandas query string.
        """
        if 'rules' in rule_object and len(rule_object['rules']) > 0:
            return self.__explain_rule_group(RuleGroup(rule_object))
        return self.__explain_rule(Rule(rule_object))

    def __explain_rule(self, rule):
        """
        Generates a pandas query string based on the given rule.

        Args:
            rule (Rule): The rule to generate the query for.

        Returns:
            str: The pandas query string.
        """
        # return f"{rule.field} {rule.operator_id} {rule.get_value()}"
        if rule.type in ['string', 'datetime', 'date']:
            return pandas_get_operator(
                f'`{self.get_property_name(rule.field)}`', rule.operator_id, f'"{rule.get_value()}"')  # pylint: disable=line-too-long
        return pandas_get_operator(
            f'`{self.get_property_name(rule.field)}`', rule.operator_id, rule.get_value())

    def __explain_rule_group(self, rule_group):
        """
        Generates a pandas query string based on the given rule group.

        Args:
            rule_group (RuleGroup): The rule group to generate the query for.

        Returns:
            str: The pandas query string.
        """
        if rule_group.condition == 'AND':
            return ' & '.join(map(self.__explain_rule_object, rule_group.rules))
        if rule_group.condition == 'OR':
            return ' | '.join(map(self.__explain_rule_object, rule_group.rules))
        raise ValueError(f"Invalid condition: {rule_group.condition}")

    def explain(self):
        """
        Generates a pandas query string based on the rules.
        """
        return f"df.query('{self.__explain_rule_group(self.rule_group)}')"



class RuleGroup:
    """
    Represents a group of rules.

    Attributes:
        condition (str): The condition for the rule group. Can be 'AND' or 'OR'.
        rules (list): The list of rules in the group.
    """

    def __init__(self, rule_group):
        self.condition = rule_group['condition']
        self.rules = rule_group['rules']



from enum import Enum

from ..context import Context
from ..gql import org_gql


class RuleScope(Enum):
    """
    Enum representing the scope of a rule (DataPart, Entity (Golden Record), Survivorship).
    """

    DATA_PART = 'DataPart'
    ENTITY = 'Entity'
    SURVIVORSHIP = 'Survivorship'


def get_rules(context: Context, scope=RuleScope.DATA_PART) -> dict:
    """
    Retrieves rules based on the specified parameters.

    Args:
            context (Context): The context object.
            scope (RuleScope, optional): The scope of the rules. Defaults to RuleScope.DATA_PART.

    Returns:
            dict: The rules data.

    """
    query = """
        query getRules($searchName: String, $isActive: Boolean, $pageNumber: Int, $sortBy: String, $sortDirection: String, $scope: String) {
            management {
                id
                rules(
                    searchName: $searchName
                    isActive: $isActive
                    pageNumber: $pageNumber
                    sortBy: $sortBy
                    sortDirection: $sortDirection
                    scope: $scope
                ) {
                    total
                    data {
                        id
                        name
                        order
                        description
                        isActive
                        createdBy
                        modifiedBy
                        ownedBy
                        createdAt
                        modifiedAt
                        author {
                            id
                            username
                            __typename
                        }
                        scope
                        isReprocessing
                        __typename
                    }
                    __typename
                }
                __typename
            }
        }
        """

    variables = {
        "scope": scope.value
    }

    return org_gql(context, query, variables)


def get_rule(context: Context, rule_id: str):
    """
    Retrieves the properties of a rule based on the provided rule ID.

    Args:
            context (Context): The context object.
            rule_id (str): The ID of the rule to retrieve properties for.

    Returns:
            dict: A dictionary containing the properties of the rule.
    """
    query = """
        query getRule($id: ID!) {
            management {
                id
                rule(id: $id) {
                    id
                    name
                    description
                    isActive
                    createdBy
                    modifiedBy
                    ownedBy
                    createdAt
                    modifiedAt
                    condition
                    actions
                    rules
                    sourceDetail {
                        id
                        name
                        type
                        __typename
                    }
                    author {
                        id
                        username
                        __typename
                    }
                    scope
                    isReprocessing
                    requiresAttention
                    __typename
                }
                __typename
            }
        }
        """
    variables = {
        "id": rule_id
    }

    return org_gql(context, query, variables)



import re


def default_get_operator(operator_id):
    """
    Retrieves the operator based on the provided operator ID.

    Args:
            operator_id (str): The ID of the operator to retrieve.

    Returns:
            dict: A dictionary containing the properties of the operator.
    """
    rule_operator_by_id = {
        # 'Is Not True'
        '7cd9d1e6-25d8-4fbf-977a-1d2c3832289d': lambda l, r, o: not is_true(l, r, o),
        # 'Is True'
        'bd804d86-f2bc-4bc1-9598-58cb6c490311': is_true,
        # 'Begins With'
        'daf3a318-569f-4baf-a17f-35b4426ae603': begins_with,
        # 'Between'
        '566bc291-df6a-427c-b7f4-00f8766d37a2': between,
        # 'Contains'
        '4988d076-3ec1-4414-9f56-5b9b30e25f72': contains,
        # 'Ends With'
        'ff9a47ac-8e3c-4123-80ee-e0b660aa5f9f': ends_with,
        # 'Equals'
        '0bafc522-8011-43ba-978a-babe222ba466': equals,
        # 'Exists'
        '306e5d29-832f-41e7-a6db-1c02eb5ebf74': lambda l, r, o: not is_null(l, r, o),
        # 'Greater'
        '85ecec6b-9c45-4365-9614-ef090137098d': greater,
        # 'Greater or Equal'
        '5a2d1185-7fd8-4dda-9d33-7f69e42e9147': greater_or_equal,
        # 'In'
        '31feed84-ce0e-435f-ba94-ce01b9c7bd15': in_list,
        # 'Is False'
        'dfbfc760-69ea-4241-b33e-0425c2a51688': lambda l, r, o: not is_true(l, r, o),
        # 'Is Not Null'
        'a42a9f87-7944-4d07-b1f3-d2f404bc4fd7': lambda l, r, o: not is_null(l, r, o),
        # 'Is Null'
        '9ad76320-021e-4211-9e9a-8b148c9c600d': is_null,
        # 'Is True'
        '04f2d73a-c908-4bc3-a7d5-1774b7db44dd': is_true,
        # 'Less'
        '0de353ce-bbe4-4a27-8e20-0dc21947e987': less,
        # 'Less or Equal'
        '42d4d22d-e37a-4c5f-96eb-7ae59513f80f': less_or_equal,
        # 'Matches pattern'
        '4e4782d7-8c17-47c1-9797-37434fce0c77': matches_pattern,
        # 'Not Begins With'
        'bbd02195-213e-40e2-9597-65ec05100dfc': lambda l, r, o: not begins_with(l, r, o),
        # 'Not Between'
        '5642150e-fa0a-462c-a230-619d97cb3f8b': lambda l, r, o: not between(l, r, o),
        # 'Not Contains'
        '4af53db0-acdb-44a6-89a1-40fa8ef05cde': lambda l, r, o: not contains(l, r, o),
        # 'Not Ends With'
        '3527ff72-bdd4-4bb4-b964-f18898d5a82d': lambda l, r, o: not ends_with(l, r, o),
        # 'Not Equal'
        '4f7f2b0f-155a-4208-b849-3e972f5d89d2': lambda l, r, o: not equals(l, r, o),
        # 'Does Not Exist'
        '16b2dc07-94e3-4c29-bc60-a7a1d25de53b': is_null,
        # 'Not In'
        '25c16d50-d170-4295-ad85-c2571431794a': lambda l, r, o: not in_list(l, r, o),
        # 'Does not match pattern'
        '9253eea0-5975-460a-8373-dd6ff57d4e65': lambda l, r, o: not matches_pattern(l, r, o)
    }

    if operator_id in rule_operator_by_id:
        return rule_operator_by_id[operator_id]

    raise ValueError(f"Operator ID '{id}' not found.")


def begins_with(left, right, _obj):
    """
    Evaluates if the left operand begins with the right operand.

    Args:
        left: The left operand.
        right: The right operand.
        _obj: Ignored parameter. (input object)

    Returns:
        True if the left operand begins with the right operand, False otherwise.
    """
    if left is None:
        return False
    return str(left).lower().startswith(str(right).lower())


def not_begins_with(left, right, _obj):
    """
    Evaluates if the left operand does not begin with the right operand.

    Args:
        left: The left operand.
        right: The right operand.
        _obj: Ignored parameter. (input object)

    Returns:
        True if the left operand does not begin with the right operand, False otherwise.
    """
    return not begins_with(left, right, _obj)


def contains(left, right, _obj):
    """
    Evaluates if the left operand contains the right operand.

    Args:
        left: The left operand.
        right: The right operand.
        _obj: Ignored parameter. (input object)

    Returns:
        True if the left operand contains the right operand, False otherwise.
    """
    if left is None:
        return False
    return str(right).lower() in str(left).lower()


def ends_with(left, right, _obj):
    """
    Evaluates if the left operand ends with the right operand.

    Args:
        left: The left operand.
        right: The right operand.
        _obj: Ignored parameter. (input object)

    Returns:
        True if the left operand ends with the right operand, False otherwise.
    """
    if left is None:
        return False
    return str(left).lower().endswith(str(right).lower())


def equals(left, right, _obj):
    """
    Evaluates if the left operand is equal to the right operand.

    Args:
        left: The left operand.
        right: The right operand.
        _obj: Ignored parameter. (input object)

    Returns:
        True if the left operand is equal to the right operand, False otherwise.
    """
    if left is None or right is None:
        return False
    return str(left.lower()) == str(right.lower())


def is_null(left, _right, _obj):
    """
    Evaluates if the given value is null.

    Args:
        left: The value to be evaluated.
        _right: Ignored parameter.
        _obj: Ignored parameter.

    Returns:
        bool: True if the value is null, False otherwise.
    """
    return left is None


def is_true(left, _right, _obj):
    """
    Check if the left operand is true.

    Args:
        left: The left operand.
        _right: The right operand (not used in this function).
        _obj: The object being evaluated (not used in this function).

    Returns:
        bool: True if the left operand is true, False otherwise.
    """
    return left


def matches_pattern(left, right, _obj):
    """
    Evaluates if the left operand matches the right operand.

    Args:
        left: The left operand.
        right: The right operand.
        _obj: Ignored parameter. (input object)

    Returns:
        True if the left operand matches the right operand, False otherwise.
    """
    if left is None:
        return False
    return bool(re.match(right, left, re.IGNORECASE))


def in_list(left, right, _obj):
    """
    Evaluates if the left operand is in the right operand.

    Args:
        left: The left operand.
        right: The right operand.
        _obj: Ignored parameter. (input object)

    Returns:
        True if the left operand is in the right operand, False otherwise.
    """
    if left is None:
        return False
    return any(str(left).lower() == str(item).lower() for item in right)


def greater(left, right, _obj):
    """
    Evaluates if the left operand is greater than the right operand.

    Args:
        left: The left operand.
        right: The right operand.
        _obj: Ignored parameter. (input object)

    Returns:
        True if the left operand is greater than the right operand, False otherwise.
    """
    if left is None or right is None:
        return False
    return left > right


def greater_or_equal(left, right, _obj):
    """
    Evaluates if the left operand is greater than or equal to the right operand.

    Args:
        left: The left operand.
        right: The right operand.
        _obj: Ignored parameter. (input object)

    Returns:
        True if the left operand is greater than or equal to the right operand, False otherwise.
    """
    if left is None or right is None:
        return False
    return left >= right


def less(left, right, _obj):
    """
    Evaluates if the left operand is less than the right operand.

    Args:
        left: The left operand.
        right: The right operand.
        _obj: Ignored parameter. (input object)

    Returns:
        True if the left operand is less than the right operand, False otherwise.
    """
    if left is None or right is None:
        return False
    return left < right


def less_or_equal(left, right, _obj):
    """
    Evaluates if the left operand is less than or equal to the right operand.

    Args:
        left: The left operand.
        right: The right operand.
        _obj: Ignored parameter. (input object)

    Returns:
        True if the left operand is less than or equal to the right operand, False otherwise.
    """
    if left is None or right is None:
        return False
    return left <= right


def between(left, right, _obj):
    """
    Evaluates if the left operand is between the right operand.

    Args:
        left: The left operand.
        right: The right operand.
        _obj: Ignored parameter. (input object)

    Returns:
        True if the left operand is between the right operand, False otherwise.
    """
    if left is None or right is None:
        return False
    return right[0] <= left <= right[1]


def pandas_get_operator(left, operator_id, right):
    """
    Retrieves the operator based on the provided operator ID.

    Args:
        left: The left operand.
        operator_id (str): The ID of the operator to retrieve.
        right: The right operand.

    Returns:
        dict: A dictionary containing the properties of the operator.
    """
    rule_operator_by_id = {
        # 'Is Not True'
        '7cd9d1e6-25d8-4fbf-977a-1d2c3832289d': lambda l, r: f"~{l}",
        # 'Is True'
        'bd804d86-f2bc-4bc1-9598-58cb6c490311': lambda l, r: l,
        # 'Begins With'
        'daf3a318-569f-4baf-a17f-35b4426ae603': lambda l, r: f"{l}.str.startswith({r})",
        # 'Between'
        '566bc291-df6a-427c-b7f4-00f8766d37a2': lambda l, r: f"({l} >= {r[0]}) & ({l} <= {r[1]})",
        # 'Contains'
        '4988d076-3ec1-4414-9f56-5b9b30e25f72': lambda l, r: f"{l}.str.contains({r})",
        # 'Ends With'
        'ff9a47ac-8e3c-4123-80ee-e0b660aa5f9f': lambda l, r: f"{l}.str.endswith({r})",
        # 'Equals'
        '0bafc522-8011-43ba-978a-babe222ba466': lambda l, r: f"{l} == {r}",
        # 'Exists'
        '306e5d29-832f-41e7-a6db-1c02eb5ebf74': lambda l, r: f"~{l}.isnull()",
        # 'Greater'
        '85ecec6b-9c45-4365-9614-ef090137098d': lambda l, r: f"{l} > {r}",
        # 'Greater or Equal'
        '5a2d1185-7fd8-4dda-9d33-7f69e42e9147': lambda l, r: f"{l} >= {r}",
        # 'In'
        '31feed84-ce0e-435f-ba94-ce01b9c7bd15': lambda l, r: f"{l}.isin({r})",
        # 'Is False'
        'dfbfc760-69ea-4241-b33e-0425c2a51688': lambda l, r: f"~{l}",
        # 'Is Not Null'
        'a42a9f87-7944-4d07-b1f3-d2f404bc4fd7': lambda l, r: f"~{l}.isnull()",
        # 'Is Null'
        '9ad76320-021e-4211-9e9a-8b148c9c600d': lambda l, r: f"{l}.isnull()",
        # 'Is True'
        '04f2d73a-c908-4bc3-a7d5-1774b7db44dd': lambda l, r: l,
        # 'Less'
        '0de353ce-bbe4-4a27-8e20-0dc21947e987': lambda l, r: f"{l} < {r}",
        # 'Less or Equal'
        '42d4d22d-e37a-4c5f-96eb-7ae59513f80f': lambda l, r: f"{l} <= {r}",
        # 'Matches pattern'
        '4e4782d7-8c17-47c1-9797-37434fce0c77': lambda l, r: f"{l}.str.match({r})",
        # 'Not Begins With'
        'bbd02195-213e-40e2-9597-65ec05100dfc': lambda l, r: f"~{l}.str.startswith({r})",
        # 'Not Between'
        '5642150e-fa0a-462c-a230-619d97cb3f8b': lambda l, r: f"({l} < {r[0]}) | ({l} > {r[1]})",
        # 'Not Contains'
        '4af53db0-acdb-44a6-89a1-40fa8ef05cde': lambda l, r: f"~{l}.str.contains({r})",
        # 'Not Ends With'
        '3527ff72-bdd4-4bb4-b964-f18898d5a82d': lambda l, r: f"~{l}.str.endswith({r})",
        # 'Not Equal'
        '4f7f2b0f-155a-4208-b849-3e972f5d89d2': lambda l, r: f"{l} != {r}",
        # 'Does Not Exist'
        '16b2dc07-94e3-4c29-bc60-a7a1d25de53b': lambda l, r: f"{l}.isnull()",
        # 'Not In'
        '25c16d50-d170-4295-ad85-c2571431794a': lambda l, r: f"~{l}.isin({r})",
        # 'Does not match pattern'
        '9253eea0-5975-460a-8373-dd6ff57d4e65': lambda l, r: f"~{l}.str.match({r})"
    }

    if operator_id in rule_operator_by_id:
        return rule_operator_by_id[operator_id](left, right)

    return f"{left} {operator_id} {right}"



import requests

from .account import get_users
from .context import Context
from .env import CLUEDIN_REQUEST_TIMEOUT_IN_SECONDS


def post_clue(context: Context, clue: str, content_type: str = 'application/xml') -> str:
    """Post a clue to CluedIn.

    Args:
        context (Context): Context object.
        clue (str): Clue XML or JSON.
        content_type (str, optional): Content type. Defaults to 'application/xml'.
          If 'application/json', the clue is expected to be JSON.

    Returns:
        str: response text.
    """
    headers = {
        'Authorization': f'Bearer {context.access_token}',
        'Content-Type': content_type
    }
    params = {'save': 'true'}

    response = requests.post(
        url=f'{context.public_api_url}/v1/clue',
        data=clue.encode(),
        headers=headers,
        params=params,
        timeout=CLUEDIN_REQUEST_TIMEOUT_IN_SECONDS,
        verify=context.verify_tls
    )

    if not response.ok:
        response.raise_for_status()

    return response.text


def restore_user_entities(context: Context) -> list:
    """Restore user entities if they are missing.
      The method gets all users for the organization and checks if they have an entity.
      If they don't, a clue is created and posted to CluedIn.

    Args:
        context (Context): Context object.

    Returns:
        list: list of results.
    """

    org_id = context.jwt_payload['OrganizationId']
    org_name = context.org_name

    results = []

    for user in get_users(context, org_id):
        if user['Entity'] is None:
            user_id = user['Account']['Id']
            user_name = user['Account']['UserName']

            # pylint: disable=line-too-long
            clue = f'''<clue organization="{org_id}" origin="/Infrastructure/User#CluedIn:{user_id}" appVersion="2.17.0.0">
          <clueDetails>
            <data inputSource="user" origin="/Infrastructure/User#CluedIn:{user_id}" appVersion="2.17.0.0">
              <processingFlags />
              <entityData id="6" origin="/Infrastructure/User#CluedIn:{user_id}" appVersion="2.17.0.0">
                <entityType>/Infrastructure/User</entityType>
                <name>{user_name}</name>
                <aliases>
                  <value>{user_id}</value>
                  <value>{user_name}</value>
                </aliases>
                <systemTagName>{user_name}</systemTagName>
                <codes>
                  <value>/Infrastructure/User#CluedIn:{user_id}</value>
                  <value>/Infrastructure/User#CluedIn:{user_name}</value>
                </codes>
                <edges>
                  <outgoing>
                    <edge type="/WorksFor" creationOptions="Default" from="C:/Infrastructure/User#CluedIn:{user_id}" to="{org_name}§C:/Organization#CluedIn:{org_id}" />
                  </outgoing>
                </edges>
                <edgesSummary />
                <properties type="/Metadata/KeyValue">
                  <property key="IsNewUser">True</property>
                  <property key="organization.name">{org_name}</property>
                  <property key="user.email">{user_name}</property>
                </properties>
              </entityData>
              <dataActions />
              <processedData ref="6" origin="/Infrastructure/User#CluedIn:{user_id}" appVersion="2.17.0.0">
                <edges />
                <edgesSummary />
                <properties type="/Metadata/KeyValue" />
                <timeToLive>0</timeToLive>
                <isShadowEntity>False</isShadowEntity>
              </processedData>
            </data>
            <references />
          </clueDetails>
        </clue>'''

            results.append({
                'user': user_name,
                'result': post_clue(context, clue)
            })
    return results



import json
import requests
from .jwt import get_jwt_payload
from .env import CLUEDIN_REQUEST_TIMEOUT_IN_SECONDS


class Context:
    """Context object. Used to store information about the CluedIn instance."""

    def __init__(self,
                 domain: str,
                 org_name: str,
                 user_email: str,
                 user_password: str,
                 protocol: str = 'https',
                 access_token: str = None,
                 org_url: str = None,
                 auth_url: str = None,
                 api_url: str = None,
                 gql_org_url: str = None,
                 gql_api_url: str = None,
                 public_api_url: str = None,
                 verify_tls: bool = True) -> None:
        self.domain = domain
        self.org_name = org_name
        self.user_email = user_email
        self.user_password = user_password
        self.protocol = protocol
        self.access_token = access_token
        if access_token:
            self.jwt_payload = get_jwt_payload(access_token)

        if org_url is None:
            self.org_url = f'{self.protocol}://{self.org_name}.{self.domain}'
        else:
            self.org_url = org_url

        if auth_url is None:
            self.auth_url = f'{self.org_url}/auth'
        else:
            self.auth_url = auth_url

        if api_url is None:
            self.api_url = f'{self.org_url}/api/api'
        else:
            self.api_url = api_url

        if gql_org_url is None:
            self.gql_org_url = f'{self.org_url}/graphql'
        else:
            self.gql_org_url = gql_org_url

        if gql_api_url is None:
            self.gql_api_url = f'{self.api_url}/graphql'
        else:
            self.gql_api_url = gql_api_url

        if public_api_url is None:
            self.public_api_url = f'{self.org_url}/public/api'
        else:
            self.public_api_url = public_api_url

        self.verify_tls = verify_tls

    @classmethod
    def from_dict(cls, context_dict: dict):
        """Create a Context object from a dictionary.

        Args:
            context_dict (dict): Dictionary containing the context information.

        Returns:
            Context: Context object.
        """
        domain: str = None
        org_name: str = None
        user_email: str = None
        user_password: str = None
        protocol: str = None
        access_token: str = None
        org_url: str = None
        auth_url: str = None
        api_url: str = None
        gql_org_url: str = None
        gql_api_url: str = None
        public_api_url: str = None
        verify_tls: bool = True

        if 'domain' in context_dict:
            domain = context_dict['domain']

        if 'org_name' in context_dict:
            org_name = context_dict['org_name']

        if 'user_email' in context_dict:
            user_email = context_dict['user_email']

        if 'user_password' in context_dict:
            user_password = context_dict['user_password']

        if 'protocol' in context_dict:
            protocol = context_dict['protocol']
        else:
            protocol = 'https'

        if 'access_token' in context_dict:
            access_token = context_dict['access_token']

        if 'org_url' in context_dict:
            org_url = context_dict['org_url']

        if 'auth_url' in context_dict:
            auth_url = context_dict['auth_url']

        if 'api_url' in context_dict:
            api_url = context_dict['api_url']

        # TODO: Remove in v3.0.0
        if 'gql_url' in context_dict:
            gql_org_url = context_dict['gql_url']

        if 'gql_org_url' in context_dict:
            gql_org_url = context_dict['gql_org_url']

        if 'gql_api_url' in context_dict:
            gql_api_url = context_dict['gql_api_url']

        if 'public_api_url' in context_dict:
            public_api_url = context_dict['public_api_url']

        if 'verify_tls' in context_dict:
            verify_tls = str(context_dict['verify_tls']).lower() == 'true'

        return cls(protocol=protocol,
                   domain=domain,
                   org_name=org_name,
                   user_email=user_email,
                   user_password=user_password,
                   access_token=access_token,
                   org_url=org_url,
                   auth_url=auth_url,
                   api_url=api_url,
                   gql_org_url=gql_org_url,
                   gql_api_url=gql_api_url,
                   public_api_url=public_api_url,
                   verify_tls=verify_tls)

    @staticmethod
    def from_json_file(file_path: str):
        """Create a Context object from a JSON file.

        Args:
            file_path (str): Path to the JSON file.

        Returns:
            Context: Context object.
        """
        with open(file_path, 'r', encoding='utf-8') as file:
            return Context.from_dict(json.load(file))

    def get_token(self) -> dict:
        """Get a new access token from the CluedIn instance.

        Returns:
            dict: Dictionary containing the access token.
        """
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded'
        }
        data = {
            'username': self.user_email,
            'password': self.user_password,
            'client_id': self.org_name,
            'grant_type': 'password'
        }
        response = requests.post(
            url=f'{self.auth_url}/connect/token',
            headers=headers,
            data=data,
            timeout=CLUEDIN_REQUEST_TIMEOUT_IN_SECONDS,
            verify=self.verify_tls)

        if not response.ok:
            response.raise_for_status()

        response_json = response.json()

        self.access_token = response_json['access_token']
        self.jwt_payload = get_jwt_payload(self.access_token)

        return response_json
